/*
 * Copyright (C) 2007-2008 Guillaume Desmottes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Guillaume Desmottes <gdesmott@gnome.org>
 */

#ifndef __TPAW_ACCOUNT_WIDGET_IRC_H__
#define __TPAW_ACCOUNT_WIDGET_IRC_H__

#include "tpaw-account-widget.h"
#include "tpaw-irc-network-chooser.h"

G_BEGIN_DECLS

TpawIrcNetworkChooser * tpaw_account_widget_irc_build (
    TpawAccountWidget *self,
    const char *filename,
    GtkWidget **table_common_settings,
    GtkWidget **box);

TpawIrcNetworkChooser * tpaw_account_widget_irc_build_simple (
    TpawAccountWidget *self,
    const char *filename,
    GtkWidget **box);

G_END_DECLS

#endif /* __TPAW_ACCOUNT_WIDGET_IRC_H__ */
